/* Copyright (c) 2006 Rice University */
/* All Rights Reserved */
/* This code is covered by the Rice-WARP license */
/* See http://warp.rice.edu/license/ for details */


#include "xmk.h"
#include <string.h>
#include <stdlib.h>
#include <os_config.h>
#include <sys/process.h>
#include <pthread.h>
#include "xparameters.h"
#include "stdio.h"
#include "xutil.h"
#include "warplib.h"
#include "xgpio_l.h"


//========================================================================
// This is a mutex example. The mutex is a more simple version of a
// semaphore, with less functionality. A semaphore can be associated with
// a specified process using sem_open() -- a mutex cannot do this. A mutex
// however, can block other process from occurring by using the same
// concept of locking.
//
// This demo will perform 10(actually closer to 12)-seconds-of-fame.
// I will generate two threads that want to share the same resource -- in
// this case, the LEDs and 7seg displays.
// The 7 segment display will be used to output which thread
// is currently running ("00" for thread A, and "11" for thread B.
// One thread will make the LEDs bounce back and forth in a Chunter-esque
// nightrider fashion. The second thread will output the LEDs in a random
// sequence.
//
// One way to think of this is that the mutex will establish the
// LEDs/7seg as a resource that must be locked on. If no mutex were
// used to lock on the resource, a jumbled mess of data would come
// out as the proc. works on them "simultaneously."
//
// This project will utilize a standard mutex -- has only one lock
// and one unlock. This is set by changing the mutext attrs
//
//
//
//
// Also note: for this example to be effective, the sub-threads must
// be of equal priority (in other words the processor will work on the
// two in Round-Robin fashion). If the threads were not of equal priority
// levels, the processor would work on one first until it finished, and
// then jump to the thread of lower priority. Lower priority threads can
// be accessed from higher-priority threads, through making joined
// pthreads.
//
// Joined pthreads causes the calling thread to be blocked/wait, until
// the called thread has completed.
//
//
// Written by:	Scott Novich (snovich@rice.edu)
// for Nokia and Rice
//
//========================================================================

// the following declarations are neccessary for thread creation
static pthread_t fame1, fame2; // an identifier for creating a new pthread
static pthread_attr_t famethreadattr; // an attribute identifer used for creating the new pthread
static struct sched_param famethreadsched; // used for giving our new thread a priority level
static pthread_mutex_t multi_resource_mutex; // declare a mutex
pthread_mutexattr_t mutex_attr; // declare mutex attribute structure
int launchcode = 0; // holder
int temp; // holder for whatever
static int temp2; // holder for whatever

void* imfamous(void *arg)
{
	int tracker1=0;
	// wait for go-code from  INITPROC
	//while(launchcode == 0)
	//{
	//	;
	//}
	while(1)
	{
		if(tracker1==0)
		{
			pthread_mutex_lock(&multi_resource_mutex);
		}

		XGpio_mSetDataReg(XPAR_LED_7SEGMENT_BASEADDR, 1,WarpLib_segmap('0'));
		XGpio_mSetDataReg(XPAR_LED_7SEGMENT_1_BASEADDR, 1,WarpLib_segmap('0'));

		XGpio_mSetDataReg(XPAR_LEDS_4BIT_BASEADDR, 1, 0x1);
		usleep(500000);// half a second, so every iteration of this loop is 3 seconds
		XGpio_mSetDataReg(XPAR_LEDS_4BIT_BASEADDR, 1, 0x2);
		usleep(500000);
		XGpio_mSetDataReg(XPAR_LEDS_4BIT_BASEADDR, 1, 0x4);
		usleep(500000);
		XGpio_mSetDataReg(XPAR_LEDS_4BIT_BASEADDR, 1, 0x8);
		usleep(500000);
		XGpio_mSetDataReg(XPAR_LEDS_4BIT_BASEADDR, 1, 0x4);
		usleep(500000);
		XGpio_mSetDataReg(XPAR_LEDS_4BIT_BASEADDR, 1, 0x2);
		usleep(500000);
		tracker1++;
		if(tracker1 == 3)
		{
			pthread_mutex_unlock(&multi_resource_mutex); // pass the access rights to the other function
			tracker1=0;
			sleep(5);
		}

	}
}

void* noimfamous(void *dummy)
{
	int tracker2=0;

	//while(launchcode == 0)
	//{
	//	;
	//}

	while(1)
	{
		if(tracker2==0)
		{
			pthread_mutex_lock(&multi_resource_mutex);
		}

		XGpio_mSetDataReg(XPAR_LED_7SEGMENT_BASEADDR, 1,WarpLib_segmap('1'));
		XGpio_mSetDataReg(XPAR_LED_7SEGMENT_1_BASEADDR, 1,WarpLib_segmap('1'));

		XGpio_mSetDataReg(XPAR_LEDS_4BIT_BASEADDR, 1, 0x8);
		usleep(500000);
		XGpio_mSetDataReg(XPAR_LEDS_4BIT_BASEADDR, 1, 0x1);
		usleep(500000);
		XGpio_mSetDataReg(XPAR_LEDS_4BIT_BASEADDR, 1, 0x4);
		usleep(500000);
		XGpio_mSetDataReg(XPAR_LEDS_4BIT_BASEADDR, 1, 0x2);
		usleep(500000);
		XGpio_mSetDataReg(XPAR_LEDS_4BIT_BASEADDR, 1, 0x8);
		usleep(500000);
		XGpio_mSetDataReg(XPAR_LEDS_4BIT_BASEADDR, 1, 0x1);
		usleep(500000);
		tracker2++;
		if(tracker2 == 3)
		{
			pthread_mutex_unlock(&multi_resource_mutex); // pass the access rights to the other function
			tracker2=0;
			sleep(5);
		}


	}

}

// the following thread is also a system initial thread. The only difference is that it's of lower
// priority.
void* INITPROC(void *alsonotused)
{
	print("\r\n -- entering INITPROC --");

	// initialize the pthreads
	pthread_attr_init(&famethreadattr);
	// pthread_attr_setdetachstate(&famethreadattr, PTHREAD_CREATE_JOINABLE);
	famethreadsched.sched_priority = 2;
	pthread_attr_setschedparam(&famethreadattr,&famethreadsched);

	print("\r\n -- Launching multi_resource_mutex --");
	// init the mutecies
	pthread_mutexattr_init(&mutex_attr); // note there are two mutex types
	pthread_mutex_init(&multi_resource_mutex, &mutex_attr);

	print("\r\n -- Launching fame-demanding threads. --");
	// start the threads
	temp = pthread_create(&fame1, &famethreadattr, (void*)imfamous, NULL);
	temp = pthread_create(&fame2, &famethreadattr, (void*)noimfamous, NULL);

	// Since this thread is priority 1, when this thread exits, the other threads will launch!

	print("\r\n -- starting the tasks! Perpetual 12-second resource switching starting now...");

	pthread_exit(&temp2);
}

int main () {
   print("-- Entering main() --\r\n");
	xilkernel_main();
	print("-- Exiting main() --\r\n");
   return 0;
}

